"""
Copyright (c) 2023, Auorui.
All rights reserved.

This module is based on PIL implemented transformations.
"""
import torch
from PIL import Image, ImageOps, ImageEnhance, ImageFilter
import random
from pyzjr.core import _check_img_is_plt, is_pil,_check_img_is_ndarray, _check_parameter_is_tuple_and_list_or_single_2, \
    _check_input_is_tensor, get_image_size, check_dtype, _get_image_num_channels
from ._utils import Images, Compose
import numpy as np

__all__ = ["PILToTensor",
           "NdarryToPIL",
           "TensorToPIL",
           "MeanStdNormalize",
           "AdjustBrightness",
           "AdjustContrast",
           "AutoContrast",
           "RandomAutoContrast",
           "AdjustGamma",
           "AdjustHue",
           "AdjustSaturation",
           "CenterCrop",
           "EqualizeHistogram",
           "RandomEqualizeHistogram",
           "RandomHorizontalFlip",
           "RandomVerticalFlip",
           "RandomPCAnoise",
           "InvertColor",
           "RandomInvertColor",
           "Resize",
           "ColorJitter",
           "RandomCrop",
           "RandomRotation",
           "Grayscale",
           "RandomGrayscale",
           "AdjustSharpness",
           "RandomAdjustSharpness",
           "GaussianBlur",
           "ResizedCrop",
           "RandomResizedCrop",
           "Pad"]


class PILToTensor():
    """
    Usage Directions:
        Convert a PIL format image to a PyTorch tensor.

    Parameters:
        - half (bool, optional): If True, the resulting tensor will be in half-precision (fp16).
          If False, the tensor will be in single-precision (fp32). Default is False.

    Examples:
        >>> pil_image = Image.open('example.jpg')  # Load a PIL image
        >>> converter = PILToTensor(half=True)
        >>> torch_tensor = converter(pil_image)

    Notes:
        This transformation is used to convert a PIL format image into a PyTorch tensor.
        It supports both single-precision (fp32) and half-precision (fp16) tensors.
    """
    def __init__(self, half=False):
        super(PILToTensor, self).__init__()
        self.half = half

    def pil_to_tensor(self, pil_img):
        img = torch.as_tensor(np.array(pil_img, copy=True))
        img = img.view(pil_img.size[1], pil_img.size[0], len(pil_img.getbands()))
        img = img.permute((2, 0, 1))  # 从HWC格式转换为CHW格式
        img = img.half() if self.half else img.float()  # 将uint8转换为fp16/32
        img /= 255.0
        return img

    def __call__(self, pil_img, *args):
        _check_img_is_plt(pil_img)
        return self.pil_to_tensor(pil_img)


class NdarryToPIL():
    """
    Usage Directions:
        Ensure that the input image is in PIL format, and it is actually a darray to PIL conversion.

    Examples:
        >>> numpy_array = np.random.rand(256, 256, 3)  # Example NumPy array with shape (H, W, C)
        >>> converter = NdarryToPIL()
        >>> pil_image = converter(numpy_array)

    Notes:
        This transformation is used to convert a NumPy array (with proper shape and data type) into a PIL format image.
    """
    def __init__(self):
        super(NdarryToPIL, self).__init__()

    def ndarry_to_pil(self, img):
        if not is_pil(img):
            _check_img_is_ndarray(img)
            if img.ndim not in (2, 3):
                raise ValueError("The dimension of input image should be 2 or 3. Got {}.".format(img.ndim))
            if img.ndim == 2:
                img = np.expand_dims(img, 2)
            if img.shape[-1] > 4:
                raise ValueError("The channel of input image should not exceed 4. Got {}.".format(img.shape[-1]))
            if img.shape[-1] == 1:
                if img.dtype not in (np.bool_, np.int8, np.int16, np.int32, np.uint8, np.uint16, np.uint32, np.float32,
                                     np.float64):
                    raise TypeError("The input image type {} is not supported when image shape is [H, W] or [H, W, 1].".format(img.dtype))
                img = img[:, :, 0]
            elif img.dtype != np.uint8:
                raise TypeError("The input image type {} is not supported when "
                                "image shape is [H, W, 2], [H, W, 3] or [H, W, 4].".format(img.dtype))
            return Image.fromarray(img)
        return img

    def __call__(self, img):
        return self.ndarry_to_pil(img)

class TensorToPIL():
    """
    Usage Directions:
        Convert a PyTorch tensor to a PIL format image.

    Examples:
        >>> tensor = torch.randn(3, 256, 256)  # Example PyTorch tensor with 3 channels and size 256x256
        >>> converter = TensorToPIL()
        >>> pil_image = converter(tensor)

    Notes:
        This transformation is used to convert a PyTorch tensor (with proper shape and data type) into a PIL format image.
    """
    def __init__(self):
        super(TensorToPIL, self).__init__()

    def tensor_to_pil(self, img):
        if not is_pil(img):
            img = img.clamp(0, 1)  # Ensure values are in the range [0, 1]
            img = img.mul(255).byte()  # Scale to [0, 255] and convert to byte
            img = img.permute(1, 2, 0)  # Rearrange dimensions from CHW to HWC
            return Image.fromarray(img.numpy())
        return img

    def __call__(self, img):
        return self.tensor_to_pil(img)


class MeanStdNormalize():
    """
    Usage Directions:
        This class is used to perform mean-std normalization on the pixel values of a PIL format image.The default is
        to use the mean and variance of Imagenet.

    Parameters:
        - mean (list or tuple): A list or tuple of mean values for each image channel.
        - std (list or tuple): A list or tuple of standard deviation values for each image channel.

    Examples:
        >>> transforms = Compose([PILToTensor(), MeanStdNormalize(), TensorToPIL()])
        >>> transformed_image = transforms(image)

    Notes:
        This transformation normalizes the pixel values of a PIL image based on specified mean and standard deviation values for each channel.
    """
    IMAGENET_MEAN = 0.485, 0.456, 0.406
    IMAGENET_STD = 0.229, 0.224, 0.225
    def __init__(self, mean=IMAGENET_MEAN, std=IMAGENET_STD):
        super(MeanStdNormalize, self).__init__()
        self.mean = mean
        self.std = std

    def mean_std_normalize(self, img):
        if img.ndim < 3:
            raise ValueError(f'Expected tensor to be a tensor image of size (..., C, H, W). Got tensor.size() = {img.size()}.')

        mean = torch.as_tensor(self.mean, dtype=img.dtype, device=img.device)
        std = torch.as_tensor(self.std, dtype=img.dtype, device=img.device)
        if (std == 0).any():
            raise ValueError('std evaluated to zero after conversion to {}, leading to division by zero.'.format(img.dtype))
        if mean.ndim == 1:
            mean = mean.view(-1, 1, 1)
        if std.ndim == 1:
            std = std.view(-1, 1, 1)
        img.sub_(mean).div_(std)
        return img

    def __call__(self, img):
        _check_input_is_tensor(img)
        return self.mean_std_normalize(img)

class AdjustBrightness():
    """
    Usage Directions:
        This class is used to adjust the brightness of PIL format images.

    Parameters:
        - brightness_factor (float): A positive number to adjust the brightness.
          Values greater than 1 increase brightness, and values between 0 and 1 decrease brightness.

    Examples:
        >>> transforms = Compose([AdjustBrightness(1.5), PILToTensor()])
        >>> transformed_image = transforms(image)
        >>> transformed_image = torch.clamp(transformed_image, 0, 1)
        >>> transformed_image = transformed_image.permute(1, 2, 0)

    Notes:
        - A high brightness_factor (greater than 1) makes the image brighter.
        - A low brightness_factor (between 0 and 1) makes the image darker.
        - Values outside the range (0, +∞) are accepted.
    """
    def __init__(self, brightness_factor):
        super(AdjustBrightness, self).__init__()
        self.brightness_factor = brightness_factor

    def _brightness(self, pil_img, brightness_factor):
        enhancer = ImageEnhance.Brightness(pil_img)
        pil_img = enhancer.enhance(brightness_factor)
        return pil_img

    def __call__(self, pil_img, *args):
        _check_img_is_plt(pil_img)
        return self._brightness(pil_img, self.brightness_factor)

class AdjustContrast():
    """
    Usage Directions :
        This class is used to adjust the contrast of a PIL image.

    Parameters:
        - contrast_factor (float): A positive number to adjust the contrast.
          Values greater than 1 increase contrast, and values between 0 and 1 decrease contrast.

    Examples :
        >>> transforms = Compose([AdjustContrast(3), PILToTensor()])
        >>> transformed_image = transforms(image)
        >>> transformed_image = torch.clamp(transformed_image, 0, 1)
        >>> transformed_image = transformed_image.permute(1, 2, 0)

    Notes:
        - High contrast (contrast_factor > 1) makes the image more vivid and enhances the difference between light and dark areas.
        - Low contrast (contrast_factor between 0 and 1) makes the image less vivid and reduces the difference between light and dark areas.
        - Values outside the range (0, +∞) are accepted.
    """
    def __init__(self, contrast_factor):
        super(AdjustContrast, self).__init__()
        self.contrast_factor = contrast_factor

    def _contrast(self, pil_img, contrast_factor):
        enhancer = ImageEnhance.Contrast(pil_img)
        img = enhancer.enhance(contrast_factor)
        return img

    def __call__(self, pil_img):
        _check_img_is_plt(pil_img)
        return self._contrast(pil_img, self.contrast_factor)

class AutoContrast():
    """
    Usage Directions:
        Automatically adjust the contrast of PIL format images.

    Parameters:
        - cutoff (float, optional): A value between 0.0 and 1.0 that controls the fraction of pixels to ignore during contrast adjustment.
          Lower values make the algorithm consider more pixels, which might lead to stronger contrast adjustment.
          Default is 0.0.

        - ignore (int, list, or None, optional): Values to ignore during the contrast adjustment process.
          This can be an integer, a list of integers, or None. Default is None.

    Examples:
        >>> transforms = Compose([AutoContrast(), PILToTensor()])
        >>> transformed_image = transforms(image)
        >>> transformed_image = torch.clamp(transformed_image, 0, 1)
        >>> transformed_image = transformed_image.permute(1, 2, 0)

    Notes:
        This transformation automatically adjusts the contrast of an image based on pixel values.
        You can set the 'cutoff' parameter to control the fraction of extreme pixel values to ignore.
    """
    def __init__(self, cutoff=0.0, ignore=None):
        super(AutoContrast, self).__init__()
        if ignore is None:
            ignore = []
        if isinstance(ignore, int):
            ignore = [ignore]
        self.cutoff = cutoff
        self.ignore = ignore

    def auto_contrast(self,img, cutoff, ignore):
        return ImageOps.autocontrast(img, cutoff, ignore)

    def __call__(self, img):
        _check_img_is_plt(img)
        return self.auto_contrast(img, self.cutoff, self.ignore)

class RandomAutoContrast(AutoContrast):
    """
    Usage Directions:
        Randomly apply the AutoContrast transformation to a PIL format image with a specified probability.

    Example:
        >>> transforms = Compose([RandomAutoContrast(pro=0.5), PILToTensor(), TensorToPIL()])
        >>> transformed_image = transforms(image)
    """
    def __init__(self, pro=0.5):
        super(RandomAutoContrast, self).__init__()
        self.pro = pro

    def __call__(self, img):
        if random.random() >= self.pro:
            return super().__call__(img,)
        return img

class AdjustGamma():
    """
    Usage Directions:
        Adjust the gamma of a PIL format image.

    Parameters:
        - gamma (float): The gamma value to adjust the image. A higher gamma (>1) makes the image brighter,
          while a lower gamma (<1) makes the image darker.
        - gain (float, optional): A gain factor to adjust the image. Default is 1.

    Examples:
        >>> transforms = Compose([AdjustGamma(3), PILToTensor(), TensorToPIL()])
        >>> transformed_image = transforms(image)
        >>> plt.imshow(transformed_image)
        >>> plt.axis('off')
        >>> plt.show()

    Notes:
        This transformation adjusts the gamma of a PIL image. A higher gamma value makes the image appear brighter,
        while a lower gamma value makes the image appear darker.
    """
    def __init__(self, gamma, gain=1):
        super(AdjustGamma, self).__init__()
        self.gamma = gamma
        self.gain = gain

    def _gamma(self, img, gamma, gain):
        gamma_table = [(255 + 1 - 1e-3) * gain * pow(x / 255., gamma) for x in range(256)]
        if len(img.split()) == 3:
            gamma_table = gamma_table * 3
            img = img.point(gamma_table)
        elif len(img.split()) == 1:
            img = img.point(gamma_table)
        return img

    def __call__(self, img):
        _check_img_is_plt(img)
        return self._gamma(img, self.gamma, self.gain)

class AdjustHue():
    """
    Usage Directions:
        Adjust the hue of the input image.

    Parameters:
        - hue_factor (float): The hue adjustment factor. Positive values shift the hue towards the red spectrum,
          while negative values shift it towards the blue-green spectrum.

    Example:
        >>> transforms = Compose([AdjustHue(-.5), PILToTensor(), TensorToPIL()])
        >>> transformed_image = transforms(image)

    Notes:
        This transformation adjusts the hue of a PIL image. The hue adjustment factor controls the direction and
        magnitude of the hue shift.
    """
    def __init__(self, hue_factor):
        super(AdjustHue, self).__init__()
        self.hue_factor = hue_factor

    def _hue(self,img, hue_factor):
        image = img
        image_hue_factor = hue_factor
        if not -0.5 <= image_hue_factor <= 0.5:
            raise ValueError('image_hue_factor {} is not in [-0.5, 0.5].'.format(image_hue_factor))
        mode = image.mode
        if mode in {'L', '1', 'I', 'F'}:
            return image
        hue, saturation, value = img.convert('HSV').split()
        np_hue = np.array(hue, dtype=np.uint8)
        with np.errstate(over='ignore'):
            np_hue += np.uint8(image_hue_factor * 255)
        hue = Image.fromarray(np_hue, 'L')
        image = Image.merge('HSV', (hue, saturation, value)).convert(mode)
        return image

    def __call__(self, img):
        _check_img_is_plt(img)
        return self._hue(img, self.hue_factor)

class AdjustSaturation():
    """
    Usage Directions:
        Adjust the saturation of a PIL format image.

    Parameters:
        - saturation_factor (float): The saturation adjustment factor. Positive values increase the saturation,
          while negative values decrease the saturation.

    Example:
        >>> transforms = Compose([AdjustSaturation(3), PILToTensor(), TensorToPIL()])
        >>> transformed_image = transforms(image)

    Notes:
        This transformation adjusts the saturation of a PIL image. The saturation adjustment factor controls the
        intensity of colors in the image.
    """
    def __init__(self, saturation_factor):
        super(AdjustSaturation, self).__init__()
        self.saturation_factor = saturation_factor

    def _saturation(self, img, saturation_factor):
        enhancer = ImageEnhance.Color(img)
        img = enhancer.enhance(saturation_factor)
        return img

    def __call__(self, img):
        _check_img_is_plt(img)
        return self._saturation(img, self.saturation_factor)


class CenterCrop():
    """
    Usage Directions:
        Center crop the image.

    Parameters:
        - size (int or tuple): The size of the desired center-cropped image. If an integer is provided, a square crop
          of that size is performed. If a tuple of (width, height) is provided, a rectangular crop with the specified
          dimensions is performed.

    Example:
        >>> transforms = Compose([AdjustSaturation(3), CenterCrop(224) , PILToTensor(), TensorToPIL()])
        >>> transformed_image = transforms(image)
        >>> print(transformed_image.size)

    Notes:
        This transformation performs center cropping on a PIL image. The center crop is obtained by selecting a
        rectangular region from the center of the input image.
    """
    def __init__(self, size):
        super(CenterCrop, self).__init__()
        if isinstance(size, int):
            size = (size, size)
        self.size = size

    def crop(self, img, x, y, width, height):
        return img.crop((x, y, x + width, y + height))

    def center_crop(self,img, size):
        if isinstance(size, int):
            size = (size, size)
        img_width, img_height = img.size
        crop_height, crop_width = size
        crop_y = int(round((img_height - crop_height) / 2.))
        crop_x = int(round((img_width - crop_width) / 2.))
        return self.crop(img, crop_x, crop_y, crop_width, crop_height)

    def __call__(self, img):
        _check_img_is_plt(img)
        return self.center_crop(img, self.size)


class EqualizeHistogram():
    """
    Usage Directions:
        Equalize the histogram of a PIL format image.

    Example:
        >>> transforms = Compose([EqualizeHistogram(), PILToTensor(), TensorToPIL()])
        >>> transformed_image = transforms(image)

    Notes:
        This transformation equalizes the histogram of a PIL image, which enhances the image's contrast.
    """
    def __init__(self):
        super(EqualizeHistogram, self).__init__()

    def _equalize(self,img):
        return ImageOps.equalize(img)

    def __call__(self, img):
        _check_img_is_plt(img)
        return self._equalize(img)

class RandomEqualizeHistogram(EqualizeHistogram):
    """
    Usage Directions:
        Randomly apply the EqualizeHistogram transformation to the input image with a specified probability.

    Example:
        >>> transforms = Compose([RandomEqualizeHistogram(0.8), PILToTensor(), TensorToPIL()])
        >>> transformed_image = transforms(image)
    """
    def __init__(self, pro=0.5):
        super(RandomEqualizeHistogram, self).__init__()
        self.pro = pro

    def __call__(self, img):
        if random.random() >= self.pro:
            return super().__call__(img)
        return img

class RandomHorizontalFlip():
    """
    Usage Directions:
        Randomly flip the input pil image horizontally.

    Example:
        >>> transforms = Compose([RandomHorizontalFlip(), PILToTensor(), TensorToPIL()])
        >>> transformed_image = transforms(image)
    """
    def __init__(self, prob=0.5):
        super(RandomHorizontalFlip, self).__init__()
        self.prob = prob

    def horizontal_flip(self,img):
        return img.transpose(Images.FLIP_LEFT_RIGHT)

    def random_horizontal_flip(self, img, prob):
        if prob > random.random():
            img = self.horizontal_flip(img)
        return img

    def __call__(self, img):
        _check_img_is_plt(img)
        return self.random_horizontal_flip(img, self.prob)

class RandomVerticalFlip():
    """
    Usage Directions:
        Randomly flip the input pil image vertically.

    Example:
        >>> transforms = Compose([RandomVerticalFlip(), PILToTensor(), TensorToPIL()])
        >>> transformed_image = transforms(image)
    """
    def __init__(self, prob=0.5):
        super(RandomVerticalFlip, self).__init__()
        self.prob = prob

    def vertical_flip(self, img):
        return img.transpose(Images.FLIP_TOP_BOTTOM)

    def random_vertical_flip(self,img, prob):
        if prob > random.random():
            img = self.vertical_flip(img)
        return img

    def __call__(self, img):
        _check_img_is_plt(img)
        return self.random_vertical_flip(img, self.prob)

class RandomPCAnoise():
    """
    Usage Directions:
        Add AlexNet-style PCA-based noise to an image.

    Parameters:
        - alpha (float): The strength of the PCA-based noise. A higher value results in stronger noise.

    Examples:
        >>> transforms = Compose([RandomPCAnoise(3), PILToTensor(), TensorToPIL()])
        >>> transformed_image = transforms(image)

    Notes:
        This transformation applies PCA-based noise to an image to introduce variations in the image's color and texture.
        The `alpha` parameter controls the strength of the noise.
    """
    def __init__(self, alpha=0.05):
        super(RandomPCAnoise, self).__init__()
        self.alpha = alpha

    def random_pca(self, img, alpha):
        if img.mode != 'RGB':
            img = img.convert("RGB")
        alpha_r = np.random.normal(loc=0.0, scale=alpha)
        alpha_g = np.random.normal(loc=0.0, scale=alpha)
        alpha_b = np.random.normal(loc=0.0, scale=alpha)
        table = np.array([
            [55.46 * -0.5675, 4.794 * 0.7192, 1.148 * 0.4009],
            [55.46 * -0.5808, 4.794 * -0.0045, 1.148 * -0.8140],
            [55.46 * -0.5836, 4.794 * -0.6948, 1.148 * 0.4203]
        ])
        pca_r = table[0][0] * alpha_r + table[0][1] * alpha_g + table[0][2] * alpha_b
        pca_g = table[1][0] * alpha_r + table[1][1] * alpha_g + table[1][2] * alpha_b
        pca_b = table[2][0] * alpha_r + table[2][1] * alpha_g + table[2][2] * alpha_b
        img_arr = np.array(img).astype(np.float64)
        img_arr[:, :, 0] += pca_r
        img_arr[:, :, 1] += pca_g
        img_arr[:, :, 2] += pca_b
        img_arr = np.uint8(np.minimum(np.maximum(img_arr, 0), 255))
        img = Image.fromarray(img_arr)
        return img

    def __call__(self, img):
        _check_img_is_plt(img)
        return self.random_pca(img, self.alpha)

class InvertColor():
    """
    Usage Directions:
        Invert the colors of a PIL format image. i.e: 255-pixel

    Examples:
        >>> transforms = Compose([InvertColor(), PILToTensor(), TensorToPIL()])
        >>> transformed_image = transforms(image)

    Notes:
        This transformation inverts the colors of a PIL image, which means it changes each color to its complementary color.
    """
    def __init__(self):
        super(InvertColor, self).__init__()

    def _invert_color(self,img):
        return ImageOps.invert(img)

    def __call__(self, img):
        _check_img_is_plt(img)
        return self._invert_color(img)


class RandomInvertColor(InvertColor):
    """
    Usage Directions:
        Randomly invert the colors of a PIL format image with a specified probability.

    Example:
        >>> transforms = Compose([RandomInvertColor(pro=0.5), PILToTensor(), TensorToPIL()])
        >>> transformed_image = transforms(image)

    """
    def __init__(self, pro=0.5):
        super(RandomInvertColor, self).__init__()
        self.pro = pro

    def __call__(self, img):
        if random.random() >= self.pro:
            return super().__call__(img)
        return img


class Resize():
    """
    Usage Directions:
        Resize the input PIL Image to the given size.

    Parameters:
        - sizes (list and tuple or int): Desired output size. If size is a sequence like (h, w), output size will be
                                        matched to this. If size is an int, smaller edge of the image will be matched
                                        to this number.i.e, if height > width, then image will be rescaled to
                                        (size * height / width, size).

    Examples:
        >>> transforms = Compose([Resize(224), PILToTensor(), TensorToPIL()])
        >>> transformed_image = transforms(image)
        >>> print(transformed_image.size)
    """
    def __init__(self, sizes):
        super(Resize, self).__init__()
        self.sizes = sizes

    def __WHratio(self, img, sizes):
        self.w, self.h = img.size
        if self.w < self.h:
            ow = sizes
            oh = int(sizes * self.h / self.w)
        else:
            oh = sizes
            ow = int(sizes * self.w / self.h)
        return (ow, oh)

    def resize(self, img, sizes, interpolation=Images.LINEAR):
        _check_parameter_is_tuple_and_list_or_single_2(sizes)
        if isinstance(sizes, int) or len(sizes) == 1:
            if (self.w <= self.h and self.w == sizes) or (self.h <= self.w and self.h == sizes):
                return img
            ratio = self.__WHratio(img, sizes)
            return img.resize(ratio, interpolation)
        else:
            return img.resize(sizes[::-1], interpolation)

    def __call__(self, img):
        _check_img_is_plt(img)
        return self.resize(img, self.sizes)

class ColorJitter():
    """
    Usage Directions:
        Adjusting the brightness, contrast, saturation, and hue of an image.

    Examples:
        >>> transforms = Compose([ColorJitter(1.2,2,1.2,0.3), PILToTensor(), TensorToPIL()])
        >>> transformed_image = transforms(image)
    """
    def __init__(self, brightness_factor, contrast_factor, saturation_factor, hue_factor):
        super(ColorJitter, self).__init__()
        self.brightness_factor = brightness_factor
        self.contrast_factor = contrast_factor
        self.saturation_factor = saturation_factor
        self.hue_factor = hue_factor

    def brightness(self, pil_img, brightness_factor):
        enhancer = ImageEnhance.Brightness(pil_img)
        pil_img = enhancer.enhance(brightness_factor)
        return pil_img

    def contrast(self, pil_img, contrast_factor):
        enhancer = ImageEnhance.Contrast(pil_img)
        img = enhancer.enhance(contrast_factor)
        return img

    def saturation(self, pil_img, saturation_factor):
        enhancer = ImageEnhance.Color(pil_img)
        img = enhancer.enhance(saturation_factor)
        return img

    def hue(self, pil_img, hue_factor):
        image = pil_img
        image_hue_factor = hue_factor
        if not -0.5 <= image_hue_factor <= 0.5:
            raise ValueError('image_hue_factor {} is not in [-0.5, 0.5].'.format(image_hue_factor))
        mode = image.mode
        if mode in {'L', '1', 'I', 'F'}:
            return image
        hue, saturation, value = pil_img.convert('HSV').split()
        np_hue = np.array(hue, dtype=np.uint8)
        with np.errstate(over='ignore'):
            np_hue += np.uint8(image_hue_factor * 255)
        hue = Image.fromarray(np_hue, 'L')
        image = Image.merge('HSV', (hue, saturation, value)).convert(mode)
        return image

    def _jitter(self, pil_img, brightness_factor, contrast_factor, saturation_factor, hue_factor):
        brightness_img = self.brightness(pil_img, brightness_factor)
        contrast_img = self.contrast(brightness_img, contrast_factor)
        saturation_img = self.saturation(contrast_img, saturation_factor)
        hue_img = self.hue(saturation_img, hue_factor)
        return hue_img

    def __call__(self, pil_img):
        _check_img_is_plt(pil_img)
        return self._jitter(pil_img, self.brightness_factor, self.contrast_factor, self.saturation_factor, self.hue_factor)

class RandomCrop():
    """
    Usage Directions:
        Random cropping on a PIL format image.

    Parameters:
        - size (int or tuple): The output size of the cropped image. If an integer is provided, the output will be a square crop.

    Examples:
        >>> transforms = Compose([RandomCrop(224), PILToTensor(), TensorToPIL()])
        >>> transformed_image = transforms(image)
        >>> print(transformed_image.size)
    """
    def __init__(self, size):
        super(RandomCrop, self).__init__()
        self.size = size

    def get_shape(self, image, output_size):
        if isinstance(output_size, int):
            output_size = [output_size, output_size]
        h, w = get_image_size(image)
        outw, outh = output_size
        if w == outw and h == outh:
            return 0, 0, h, w

        if h + 1 < outh or w + 1 < outw:
            raise ValueError(f"Required crop size {(outh, outw)} is larger then input image size {(h, w)}")
        i = torch.randint(0, h - outh + 1, size=(1, )).item()
        j = torch.randint(0, w - outw + 1, size=(1, )).item()
        return i, j, outh, outw

    def crop(self, img, x, y, width, height):
        return img.crop((x, y, x + width, y + height))

    def __call__(self, image):
        _check_img_is_plt(image)
        i, j, h, w = self.get_shape(image, self.size)
        return self.crop(image, j, i, w, h)  # 注意这里坐标的顺序


class RandomRotation():
    """
    Usage Directions:
        Rotate the input PIL Image by a random angle.

    Examples:
        >>> transforms = Compose([RandomRotation(), PILToTensor(), TensorToPIL()])
        >>> transformed_image = transforms(image)

    Notes:
        This transformation rotates the input PIL image by a random angle selected from [0°, 90°, 180°, 270°].
    """
    def __init__(self):
        super(RandomRotation, self).__init__()

    def rotate(self, img, angle):
        rotated_img = img.rotate(angle)
        return rotated_img

    def __call__(self, img):
        angle = random.choice([0, 90, 180, 270])
        return self.rotate(img, angle)

class Grayscale():
    """
    Convert a PIL format image to grayscale.

    Example:
        >>> transforms = Compose([Grayscale(), PILToTensor(), TensorToPIL()])
        >>> transformed_image = transforms(image)
    """
    def __init__(self):
        super(Grayscale, self).__init__()

    def to_gray(self, img):
        channels = _get_image_num_channels(img)
        if channels == 1:
            img = img.convert('L')
        elif channels == 3:
            img = img.convert('L')
            np_img = np.array(img, dtype=np.uint8)
            np_img = np.dstack([np_img, np_img, np_img])
            img = Image.fromarray(np_img, 'RGB')
        return img

    def __call__(self, img):
        _check_img_is_plt(img)
        return self.to_gray(img)

class RandomGrayscale(Grayscale):
    """
    Usage Directions:
        Randomly apply the Grayscale transformation to a PIL format image with a specified probability.

    Example:
        >>> transforms = Compose([RandomGrayscale(pro=0.5), PILToTensor(), TensorToPIL()])
        >>> transformed_image = transforms(image)
    """
    def __init__(self, pro=0.5):
        super(RandomGrayscale, self).__init__()
        self.pro = pro

    def __call__(self, img):
        if random.random() >= self.pro:
            return super().__call__(img)
        return img

class AdjustSharpness():
    """
    Usage Directions:
        Adjust the sharpness of a PIL format image.

    Parameters:
        sharpness_factor (float): The factor to adjust sharpness. A value less than 1.0 decreases sharpness, while a
        value greater than 1.0 increases sharpness. 1.0 represents the original sharpness.

    Example:
        >>> transforms = Compose([AdjustSharpness(2.0), PILToTensor(), TensorToPIL()])
        >>> transformed_image = transforms(image)
    """
    def __init__(self, sharpness_factor):
        super(AdjustSharpness, self).__init__()
        self.sharpness_factor = sharpness_factor

    def adjust_sharpness(self, img, sharpness_factor):
        _check_img_is_plt(img)
        enhancer = ImageEnhance.Sharpness(img)
        img = enhancer.enhance(sharpness_factor)
        return img

    def __call__(self, img):
        return self.adjust_sharpness(img,self.sharpness_factor)

class RandomAdjustSharpness(AdjustSharpness):
    """
    Usage Directions:
        Randomly adjust the clarity of the input image
    Example:
        >>> transforms = Compose([RandomAdjustSharpness(4.0), PILToTensor(), TensorToPIL()])
        >>> transformed_image = transforms(image)
    """
    def __init__(self, sharpness_factor, pro=0.5):
        super(RandomAdjustSharpness, self).__init__(sharpness_factor)  # 修正super()参数，调用父类的__init__
        self.sharpness_factor = sharpness_factor
        self.pro = pro

    def __call__(self, img):
        if random.random() >= self.pro:
            return super().__call__(img)
        return img

class GaussianBlur():
    """
    Usage Directions:
        Apply Gaussian blur to a PIL format image.

    Parameters:
        - radius (int): The radius of the Gaussian blur kernel. A larger value results in a stronger blur effect.

    Example:
        >>> transforms = Compose([GaussianBlur(2), PILToTensor(), TensorToPIL()])
        >>> transformed_image = transforms(image)
    """
    def __init__(self, radius):
        self.radius = radius

    def apply_gaussian_blur(self, img, radius):
        type = check_dtype(radius)
        if type.is_int():
            blurred_img = img.filter(ImageFilter.GaussianBlur(radius))
            return blurred_img

    def __call__(self, img):
        _check_img_is_plt(img)
        return self.apply_gaussian_blur(img, self.radius)


class Pad():
    """
    Usage Directions:
        Apply padding to a PIL format image.

    Parameters:
        - padding (int, tuple, or list): The padding to be applied. It can be an integer for uniform padding or a tuple/list
          of size 2 or 4 for specifying padding on each side separately.
        - padding_mode (str): The padding mode, which should be one of 'constant', 'edge', 'reflect', or 'symmetric'.
        - fill_value (tuple): The fill value used when padding_mode is 'constant'. This should be a tuple specifying the fill
          color in the format (R, G, B).

    Examples:
        >>> transforms = Compose([Pad(10), PILToTensor(), TensorToPIL()])
        >>> transformed_image = transforms(image)
    """
    def __init__(self, padding, padding_mode='constant',fill_value=(128, 128, 128)):
        super(Pad, self).__init__()
        self.padding = padding
        self.padding_mode = padding_mode
        self.fill_value = fill_value
        if self.padding_mode not in ['constant', 'edge', 'reflect', 'symmetric']:
            raise ValueError("Padding mode should be 'constant', 'edge', 'reflect', or 'symmetric'.")

    def pad(self, img, padding, padding_mode, fill_value):
        type = check_dtype(padding)
        if type.is_int():
            top = bottom = left = right = padding

        elif type.is_list_or_tuple():
            if len(padding) == 2:
                left = right = padding[0]
                top = bottom = padding[1]
            elif len(padding) == 4:
                left, top, right, bottom = padding
            else:
                raise ValueError("The size of the padding list or tuple should be 2 or 4.")
        else:
            raise TypeError("Padding can be any of: a number, a tuple or list of size 2 or 4.")
        if padding_mode == 'constant':
            if img.mode == 'P':
                palette = img.getpalette()
                image = ImageOps.expand(img, border=(left, top, right, bottom), fill=fill_value)
                image.putpalette(palette)
                return image
            if isinstance(fill_value, tuple) and (img.mode == 'L' or img.mode == '1'):
                fill_value = (fill_value[0],)
            return ImageOps.expand(img, border=(left, top, right, bottom), fill=fill_value)
        if img.mode == 'P':
            palette = img.getpalette()
            img = np.asarray(img)
            img = np.pad(img, ((top, bottom), (left, right)), padding_mode)
            img = Image.fromarray(img)
            img.putpalette(palette)
            return img
        img = np.asarray(img)
        if len(img.shape) == 3:
            img = np.pad(img, ((top, bottom), (left, right), (0, 0)), padding_mode)
        if len(img.shape) == 2:
            img = np.pad(img, ((top, bottom), (left, right)), padding_mode)
        return Image.fromarray(img)

    def __call__(self, pil_img):
        _check_img_is_plt(pil_img)
        return self.pad(pil_img,self.padding,self.padding_mode,self.fill_value)

class ResizedCrop(Resize):
    """
    Usage Directions:
        Crop and resize a PIL format image to the specified size.

    Parameters:
        - sizes (tuple): The target size for resizing in the format (width, height).
        - box (tuple): A tuple in the format (x, y, width, height) specifying the crop box coordinates.

    Examples:
        >>> transforms = Compose([ResizedCrop(224, (10, 10, 100, 100)), PILToTensor(), TensorToPIL()])
        >>> transformed_image = transforms(image)
    """
    def __init__(self, sizes, box):
        super(ResizedCrop, self).__init__(sizes)
        type = check_dtype(sizes)
        if type.is_int():
            self.sizes = [sizes, sizes]
        self.box = box

    def crop_img(self, img, x, y, width, height):
        return img.crop((x, y, x + width, y + height))

    def resize_img(self, img, size):
        return super().resize(img, size)

    def resize_crop(self, img, x, y, width, height, size):
        img = self.crop_img(img, x, y, width, height)
        img = self.resize_img(img, size)
        return img

    def __call__(self, img):
        _check_img_is_plt(img)
        if len(self.box) == 4:
            x, y, width, height = self.box
            return self.resize_crop(img, x, y, width, height, self.sizes)
        else:
            raise ValueError("Box should be a tuple of 4 integers (x, y, width, height).")


class RandomResizedCrop(ResizedCrop):
    """
    Usage Directions:
        Randomly crop and resize a PIL format image.

    Parameters:
        - sizes (tuple): The target size for resizing in the format (width, height).
        - scale (tuple): A tuple (min_scale, max_scale) specifying the range for random scaling.
        - box (tuple, optional): A tuple (x, y, width, height) specifying a fixed crop box. If None, a random box will be generated.

    Examples:
        >>> transforms = Compose([RandomResizedCrop(224, (0.2, 1.0)), PILToTensor(), TensorToPIL()])
        >>> transformed_image = transforms(image)
    """
    def __init__(self, sizes, scale=(0.2, 1.0), box=None):
        super(RandomResizedCrop, self).__init__(sizes,box)
        self.scale = scale
        self.box = box

    def get_random_box(self, image, scale):
        width, height = image.size
        minscale_img = (width * scale[0], height * scale[0])
        maxscale_img = (width * scale[1], height * scale[1])
        w = random.randint(minscale_img[0], maxscale_img[0])
        h = random.randint(minscale_img[1], maxscale_img[1])
        if w <= width and h <= height:
            x = random.randint(0, width - w)
            y = random.randint(0, height - h)
            return x, y, w, h
        else:
            return 0, 0, width, height

    def __call__(self, img):
        _check_img_is_plt(img)
        if self.box is None:
            x, y, w, h = self.get_random_box(img, self.scale)
        else:
            x, y, w, h = self.box
        return super().resize_crop(img, x, y, w, h, self.sizes)

if __name__ == "__main__":
    import matplotlib
    import matplotlib.pyplot as plt

    matplotlib.use("TkAgg")
    image = Image.open(r'D:\PythonProject\Torchproject\classification\dataset\crack\1181.png')

    transforms = Compose([RandomResizedCrop(224), PILToTensor(), TensorToPIL()])
    transformed_image = transforms(image)

    print(transformed_image.size)
    # transformed_image = torch.clamp(transformed_image, 0, 1)
    # transformed_image = transformed_image.permute(1, 2, 0)

    plt.imshow(transformed_image)
    plt.axis('off')
    plt.show()



import cv2
import numpy as np
import pyzjr.Z as Z
from PIL import Image

def RemoveInboxes(boxes):
    """
    中心点的方法,去除重叠的框,并去掉相对较小的框
    :param boxes: [minc, minr, maxc, maxr]
    :return: 过滤的boxes
    """
    final_boxes = []
    for box in boxes:
        x1, y1, x2, y2 = box
        cx = (x1 + x2) / 2
        cy = (y1 + y2) / 2
        box_area = (x2 - x1) * (y2 - y1)
        is_inner = False
        for fb in final_boxes:
            fx1, fy1, fx2, fy2 = fb
            if fx1 <= cx <= fx2 and fy1 <= cy <= fy2:
                fb_area = (fx2 - fx1) * (fy2 - fy1)
                if box_area >= fb_area:
                    final_boxes.remove(fb)
                else:
                    is_inner = True
                    break
        if not is_inner:
            final_boxes.append(box)
    return final_boxes

def cvt8png(pngpath, bit_depth=False, target=Z.white, convert=(128, 0, 0)):
    """
    Voc: RGB png彩图转换
    :param bit_depth: 默认转为8位,需要使用out_png.save可以正确保存
    True:                      False:
        plt.imshow(img)             cv2.imshow("img",img)
        plt.axis('off')             cv2.waitKey(0)
        plt.show()
    :param pngpath: 为保证是按照cv2的方式读入,所以传入路径即可
    :param target: 目标颜色,RGB方式
    :param convert: 转换颜色,同RGB格式
    :return:
    """
    png = cv2.imread(pngpath)
    png = cv2.cvtColor(png, cv2.COLOR_BGR2RGB)
    h, w = png.shape[:2]

    mask = np.all(png == target, axis=-1)
    out_png = np.zeros((h, w, 3), dtype=np.uint8)
    out_png[mask] = convert
    out_png[~mask] = png[~mask]

    if bit_depth:
        out_png_pil = Image.fromarray(out_png)
        out_png_pil = out_png_pil.convert("P", palette=Image.ADAPTIVE, colors=256)
        return out_png_pil
    else:
        out_png_cv = cv2.cvtColor(out_png, cv2.COLOR_RGB2BGR)
        return out_png_cv

def cvtMask(pngpath, lower ,upper, convert=(255, 255, 255)):
    """
    去除标签中的杂色,并对其进行颜色的转换

    :param pngpath: 输入图像的文件路径
    :param lower: 颜色范围的下限值，作为一个包含(B, G, R)值的元组
    :param upper: 颜色范围的上限值，作为一个包含(B, G, R)值的元组
    :param convert: 需要转换为的颜色，作为一个包含(B, G, R)值的元组
    :return: 处理后的图像
    """
    image = cv2.imread(pngpath)
    lower = np.array(lower, dtype=np.uint8)
    upper = np.array(upper, dtype=np.uint8)

    mask = cv2.inRange(image, lower, upper)
    image[mask > 0] = convert

    return image

def imageContentMove(image, H_pixels=None, V_pixels=None):
    if H_pixels is None:
        H_pixels = 0
    if V_pixels is None:
        V_pixels = 0
    h, w = image.shape[:2]
    black = np.zeros_like(image)

    start_row, end_row = max(0, V_pixels), min(h, h + V_pixels)
    start_col, end_col = max(0, H_pixels), min(w, w + H_pixels)

    fill_start_row, fill_end_row = max(0, -V_pixels), min(h, h - V_pixels)
    fill_start_col, fill_end_col = max(0, -H_pixels), min(w, w - H_pixels)

    black[start_row:end_row, start_col:end_col] = image[fill_start_row:fill_end_row, fill_start_col:fill_end_col]

    return black




import torch
from collections import defaultdict, deque
import time
import datetime
from pyzjr.dlearn.tools import LoadingBar

class SmoothedValue(object):
    """Track a series of values and provide access to smoothed values over a
    window or the global series average.
    """
    def __init__(self, window_size=20, fmt=None):
        if fmt is None:
            fmt = "{value:.4f} ({global_avg:.4f})"
        self.deque = deque(maxlen=window_size)
        self.total = 0.0
        self.count = 0
        self.fmt = fmt

    def update(self, value, n=1):
        self.deque.append(value)
        self.count += n
        self.total += value * n

    @property
    def median(self):
        d = torch.tensor(list(self.deque))
        return d.median().item()

    @property
    def avg(self):
        d = torch.tensor(list(self.deque), dtype=torch.float32)
        return d.mean().item()

    @property
    def global_avg(self):
        return self.total / self.count

    @property
    def max(self):
        return max(self.deque)

    @property
    def value(self):
        return self.deque[-1]

    def __str__(self):
        return self.fmt.format(
            median=self.median,
            avg=self.avg,
            global_avg=self.global_avg,
            max=self.max,
            value=self.value)

class MetricLogger(object):
    def __init__(self, delimiter="\t", load_bar = False):
        self.meters = defaultdict(SmoothedValue)
        self.delimiter = delimiter

    def update(self, **kwargs):
        for k, v in kwargs.items():
            if isinstance(v, torch.Tensor):
                v = v.item()
            assert isinstance(v, (float, int))
            self.meters[k].update(v)

    def __getattr__(self, attr):
        if attr in self.meters:
            return self.meters[attr]
        if attr in self.__dict__:
            return self.__dict__[attr]
        raise AttributeError("'{}' object has no attribute '{}'".format(
            type(self).__name__, attr))

    def __str__(self):
        loss_str = []
        for name, meter in self.meters.items():
            loss_str.append(
                "{}: {}".format(name, str(meter))
            )
        return self.delimiter.join(loss_str)

    def add_meter(self, name, meter):
        self.meters[name] = meter

    def log_every(self, iterable, print_freq, header=None):
        i = 0
        if not header:
            header = ''
        start_time = time.time()
        end = time.time()
        iter_time = SmoothedValue(fmt='{avg:.4f}')
        data_time = SmoothedValue(fmt='{avg:.4f}')
        space_fmt = ':' + str(len(str(len(iterable)))) + 'd'

        load_bar = LoadingBar(20)

        if torch.cuda.is_available():
            log_msg = self.delimiter.join([
                header,
                '[{0' + space_fmt + '}/{1}]',
                'Process: {Process}',
                'eta: {eta}',
                '{meters}',
                'time: {time}',
                'data: {data}',
                'max mem: {memory:.0f}'
            ])
        else:
            log_msg = self.delimiter.join([
                header,
                '[{0' + space_fmt + '}/{1}]',
                'Process: {Process}',
                'eta: {eta}',
                '{meters}',
                'time: {time}',
                'data: {data}'
            ])
        MB = 1024.0 * 1024.0

        for obj in iterable:
            data_time.update(time.time() - end)
            yield obj
            iter_time.update(time.time() - end)
            if i % print_freq == 0:
                eta_seconds = iter_time.global_avg * (len(iterable) - i)
                eta_string = str(datetime.timedelta(seconds=int(eta_seconds)))
                if torch.cuda.is_available():
                    progress = (i + 1) / len(iterable)
                    bar_string = load_bar(progress)
                    print("\r", log_msg.format(
                        i,
                        len(iterable),
                        Process=bar_string,
                        eta=eta_string,
                        meters=str(self),
                        time=str(iter_time), data=str(data_time),
                        memory=torch.cuda.max_memory_allocated() / MB), end=' ', flush=True)
                else:
                    progress = (i + 1) / len(iterable)
                    bar_string = load_bar(progress)
                    print("\r", log_msg.format(
                        i, bar_string,
                        len(iterable),
                        eta=eta_string,
                        meters=str(self),
                        time=str(iter_time), data=str(data_time)), end=' ', flush=True)
            i += 1
            end = time.time()
        total_time = time.time() - start_time
        total_time_str = str(datetime.timedelta(seconds=int(total_time)))
        print("\n", f'{header} Total time: {total_time_str}')

class ConfusionMatrix(object):
    """For details:https://blog.csdn.net/m0_62919535/article/details/132893016"""
    def __init__(self, num_classes):
        self.num_classes = num_classes
        self.mat = None

    def update(self, true, pred):
        t, p = true.flatten(), pred.flatten()
        n = self.num_classes
        if self.mat is None:
            self.mat = torch.zeros((n, n), dtype=torch.int64, device=t.device)
        with torch.no_grad():
            k = (t >= 0) & (t < n)
            inds = n * t[k].to(torch.int64) + p[k]
            self.mat += torch.bincount(inds, minlength=n**2).reshape(n, n)

    def reset(self):
        if self.mat is not None:
            self.mat.zero_()

    @property
    def ravel(self):
        """
        计算混淆矩阵的TN, FP, FN, TP
        支持二分类和多分类
        """
        h = self.mat.float()
        n = self.num_classes
        if n == 2:
            TP, FN, FP, TN = h.flatten()
            return TP, FN, FP, TN
        if n > 2:
            TP = h.diag()
            FN = h.sum(dim=1) - TP
            FP = h.sum(dim=0) - TP
            TN = torch.sum(h) - (torch.sum(h, dim=0) + torch.sum(h, dim=1) - TP)

            return TP, FN, FP, TN

    def compute(self):
        """
        主要在eval的时候使用,你可以调用ravel获得TN, FP, FN, TP, 进行其他指标的计算
        计算全局预测准确率(混淆矩阵的对角线为预测正确的个数)
        计算每个类别的准确率
        计算每个类别预测与真实目标的iou,IoU = TP / (TP + FP + FN)
        """
        h = self.mat.float()
        acc_global = torch.diag(h).sum() / h.sum()
        acc = torch.diag(h) / h.sum(1)
        iu = torch.diag(h) / (h.sum(1) + h.sum(0) - torch.diag(h))
        return acc_global, acc, iu

    def __str__(self):
        acc_global, acc, iu = self.compute()
        return (
            'global correct: {:.1f}\n'
            'average row correct: {}\n'
            'IoU: {}\n'
            'mean IoU: {:.1f}').format(
            acc_global.item() * 100,
            ['{:.1f}'.format(i) for i in (acc * 100).tolist()],
            ['{:.1f}'.format(i) for i in (iu * 100).tolist()],
            iu.mean().item() * 100)

class NumericalMeter:
    """Computes and stores the average and current value"""
    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count






# --------------------------------------- #
# Used for some operations on files,
# file management is abbreviated as FM.
# --------------------------------------- #
import os
import datetime
import csv
from os import getcwd
import win32com.client, gc
from datetime import datetime
from pathlib import Path
import logging
import re

__all__=[
    # def
    "get_logger",
    "getFilePath",
    "mkdir",
    "logdir",
    "Data2csv",
    "is_url",
    "is_ascii",
    "file_age",
    "file_date",
    "file_size",
    # class
    "Microsoft2PDF",
]

def get_logger(filename='./exp.log', verbosity=1, name=None):
    """示例
    logger.debug('This is a DEBUG message.')  # DEBUG 级别的日志
    logger.info('This is an INFO message.')    # INFO 级别的日志
    logger.warning('This is a WARNING message.')  # WARNING 级别的日志
    try:
        result = 10 / 0
    except Exception as e:
        logger.error(f'An error occurred: {str(e)}', exc_info=True)
    logger.critical('This is a CRITICAL message.')  # CRITICAL 级别的日志
    """
    level_dict = {0: logging.DEBUG, 1: logging.INFO, 2: logging.WARNING}
    formatter = logging.Formatter(
        "[%(asctime)s][%(filename)s]%(message)s"
    )
    logger = logging.getLogger(name)
    logger.setLevel(level_dict[verbosity])

    fh = logging.FileHandler(filename, "w")
    fh.setFormatter(formatter)
    logger.addHandler(fh)

    sh = logging.StreamHandler()
    sh.setFormatter(formatter)
    logger.addHandler(sh)

    return logger

def getFilePath(filedir,format='png'):
    """What is returned is a list that includes all paths under the target path that match the suffix."""
    file_list = [os.path.join(root, filespath) \
                 for root, dirs, files in os.walk(filedir) \
                 for filespath in files \
                 if str(filespath).endswith(format)
                 ]
    return file_list if file_list else []

def mkdir(path, *paths, inc=False):
    """
    如果构建的输出目录路径不存在，它将创建该目录。
    如果 inc 为 True，它将检查是否已经存在相同的目录。如果存在，将在目录名称末尾附加一个数字后缀，直到找到一个不存在的目录。
    返回最终的输出目录路径。
    """
    outdir = os.path.join(path, *paths)
    if not os.path.exists(outdir):
        os.makedirs(outdir)
    elif inc:
        count = 1
        outdir_inc = outdir + '_' + str(count)
        while os.path.exists(outdir_inc):
            count = count + 1
            outdir_inc = outdir + '_' + str(count)
            assert count < 100
        outdir = outdir_inc
        os.makedirs(outdir)
    return outdir


def logdir(dir="logs", format=True, prefix="", suffix=""):
    """
    Logging generator
    :param dir: Default "logs"
    :param format: DATE_FORMAT = '%A_%d_%B_%Y_%Hh_%Mm_%Ss'
                   SIMPLEDATE_FORMAT = '%Y_%m_%d_%H_%M_%S'
    :param prefix: Add folder name before time string
    :param suffix: After adding the folder name to the time string
    :return:
    """
    DATE_FORMAT = '%A_%d_%B_%Y_%Hh_%Mm_%Ss'
    SIMPLEDATE_FORMAT = '%Y_%m_%d_%H_%M_%S'
    formats = DATE_FORMAT if format else SIMPLEDATE_FORMAT
    time_str = datetime.now().strftime(formats)

    folder_names = [prefix, time_str, suffix]
    folder_names = [folder for folder in folder_names if folder]

    log_dir = os.path.join(dir, *folder_names)
    os.makedirs(log_dir, exist_ok=True)

    return log_dir

def Data2csv(header, value, log_dir, savefile_name):
    """Export data to CSV format
    Args:
        header (list): 列的标题
        value (list): 对应列的值
        log_dir (str): 文件夹路径
        savefile_name (str): 文件名（包括路径）
    """
    os.makedirs(log_dir, exist_ok=True)
    file_existence = os.path.isfile(savefile_name)

    if not file_existence:
        with open(savefile_name, 'w', newline='') as file:
            writer = csv.writer(file)
            writer.writerow(header)
            writer.writerow(value)
    else:
        with open(savefile_name, 'a', newline='') as file:
            writer = csv.writer(file)
            writer.writerow(value)

class Microsoft2PDF():
    """Convert Microsoft Office documents (Word, Excel, PowerPoint) to PDF format"""
    def __init__(self,filePath = ""):
        """
        :param filePath: If the default is a null character, it defaults to the current path
        """
        self.flagW = self.flagE = self.flagP = 1
        self.words = []
        self.ppts = []
        self.excels = []

        if filePath == "":
            filePath = getcwd()
        folder = filePath + '\\pdf\\'
        os.makedirs(folder, exist_ok=True)
        self.folder = folder

        self.filePath = filePath
        for i in os.listdir(self.filePath):
            if i.endswith(('.doc', 'docx')):
                self.words.append(i)
            if i.endswith(('.ppt', 'pptx')):
                self.ppts.append(i)
            if i.endswith(('.xls', 'xlsx')):
                self.excels.append(i)

        if len(self.words) < 1:
            print("\n[pyzjr]:No Word files\n")
            self.flagW = 0
        if len(self.ppts) < 1:
            print("\n[pyzjr]:No PPT file\n")
            self.flagE = 0
        if len(self.excels) < 1:
            print("\n[pyzjr]:No Excel file\n")
            self.flagP = 0

    def Word2Pdf(self):
        if self.flagW == 0:
            return 0
        else:
            print("\n[Start Word -> PDF conversion]")
            try:
                print("Open Word Process...")
                word = win32com.client.Dispatch("Word.Application")
                word.Visible = 0
                word.DisplayAlerts = False
                doc = None
                for i in range(len(self.words)):
                    print(i)
                    fileName = self.words[i]  # file name
                    fromFile = os.path.join(self.filePath, fileName)  # file address
                    toFileName = self.changeSufix2Pdf(fileName)  # Generated file name
                    toFile = self.toFileJoin(toFileName)  # Generated file address

                    print("Conversion:" + fileName + "in files...")
                    try:
                        doc = word.Documents.Open(fromFile)
                        doc.SaveAs(toFile, 17)
                        print("Convert to:" + toFileName + "file completion")
                    except Exception as e:
                        print(e)

                print("All Word files have been printed")
                print("End Word Process...\n")
                doc.Close()
                doc = None
                word.Quit()
                word = None
            except Exception as e:
                print(e)
            finally:
                gc.collect()

    def Excel2Pdf(self):
        if self.flagE == 0:
            return 0
        else:
            print("\n[Start Excel -> PDF conversion]")
            try:
                print("open Excel Process...")
                excel = win32com.client.Dispatch("Excel.Application")
                excel.Visible = 0
                excel.DisplayAlerts = False
                wb = None
                ws = None
                for i in range(len(self.excels)):
                    print(i)
                    fileName = self.excels[i]
                    fromFile = os.path.join(self.filePath, fileName)

                    print("Conversion:" + fileName + "in files...")
                    try:
                        wb = excel.Workbooks.Open(fromFile)
                        for j in range(wb.Worksheets.Count):  # Number of worksheets, one workbook may have multiple worksheets
                            toFileName = self.addWorksheetsOrder(fileName, j + 1)
                            toFile = self.toFileJoin(toFileName)

                            ws = wb.Worksheets(j + 1)
                            ws.ExportAsFixedFormat(0, toFile)
                            print("Convert to:" + toFileName + "file completion")
                    except Exception as e:
                        print(e)
                print("All Excel files have been printed")
                print("Ending Excel process...\n")
                ws = None
                wb.Close()
                wb = None
                excel.Quit()
                excel = None
            except Exception as e:
                print(e)
            finally:
                gc.collect()

    def PPt2Pdf(self):
        if self.flagP == 0:
            return 0
        else:
            print("\n[Start PPT -> PDF conversion]")
            try:
                print("Open PowerPoint process...")
                powerpoint = win32com.client.Dispatch("PowerPoint.Application")
                ppt = None
                for i in range(len(self.ppts)):
                    print(i)
                    fileName = self.ppts[i]
                    fromFile = os.path.join(self.filePath, fileName)
                    toFileName = self.changeSufix2Pdf(fileName)
                    toFile = self.toFileJoin(toFileName)

                    print("Conversion:" + fileName + "in files...")
                    try:
                        ppt = powerpoint.Presentations.Open(fromFile, WithWindow=False)
                        if ppt.Slides.Count > 0:
                            ppt.SaveAs(toFile, 32)
                            print("Convert to:" + toFileName + "file completion")
                        else:
                            print("Error, unexpected: This file is empty, skipping this file")
                    except Exception as e:
                        print(e)
                print("All PPT files have been printed")
                print("Ending PowerPoint process...\n")
                ppt.Close()
                ppt = None
                powerpoint.Quit()
                powerpoint = None
            except Exception as e:
                print(e)
            finally:
                gc.collect()

    def WEP2Pdf(self):
        """
        Word, Excel and PPt are all converted to PDF.
        If there are many files, it may take some time
        """
        print("Convert Microsoft Three Musketeers to PDF")
        self.Word2Pdf()
        self.Excel2Pdf()
        self.PPt2Pdf()
        print(f"All files have been converted, you can find them in the {self.folder}")

    def changeSufix2Pdf(self,file):
        """Change the file suffix to .pdf"""
        return file[:file.rfind('.')] + ".pdf"

    def addWorksheetsOrder(self,file, i):
        """Add worksheet order to file name"""
        return file[:file.rfind('.')] + "_worksheet" + str(i) + ".pdf"

    def toFileJoin(self, file):
        """Connect the file path and file name to the complete file path"""
        return os.path.join(self.filePath, 'pdf', file[:file.rfind('.')] + ".pdf")

URL_REGEX = re.compile(r'http://|https://|ftp://|file://|file:\\')

def is_url(filename):
    """Return True if string is an http or ftp path."""
    return (isinstance(filename, str) and
            URL_REGEX.match(filename) is not None)

def is_ascii(s):
    """
    Check if the string is composed of only ASCII characters.
    Args:
        s (str): String to be checked.
    Returns:
        bool: True if the string is composed only of ASCII characters, False otherwise.
    """
    s = str(s)
    return all(ord(c) < 128 for c in s)

def file_age(path, detail=False):
    """Returns the number of days since the last file update."""
    dt = (datetime.now() - datetime.fromtimestamp(Path(path).stat().st_mtime))  # delta
    se = 0
    if detail:
        se = dt.seconds / 86400
    return f"{dt.days + se} days"

def file_date(path=__file__):
    """Return readable file modification date,i.e:'2021-3-26'."""
    t = datetime.fromtimestamp(Path(path).stat().st_mtime)
    return f'{t.year}-{t.month}-{t.day}'

def file_size(path):
    """Return file/dir size (MB)."""
    if isinstance(path, (str, Path)):
        mb = 1 << 20  # bytes -> MiB (1024 ** 2)
        path = Path(path)
        if path.is_file():
            size_mb = path.stat().st_size / mb
            return f"{size_mb:.5f} MB"
        elif path.is_dir():
            total_size = sum(f.stat().st_size for f in path.glob('**/*') if f.is_file())
            size_mb = total_size / mb
            return f"{size_mb:.5f} MB"
    return "0.0 MB"


"""
Copyright (c) 2023, Auorui.
All rights reserved.

The following citation imports roughly follow the sorting by year
"""
# 1990s
from .lenet import LeNet

# 2010s
from .alexnet import AlexNet
from .vgg import VGG, vgg11_bn, vgg13_bn, vgg16_bn, vgg19_bn
from .googlenet import GoogLeNet
from .resnet import ResNet, resnet18, resnet34, resnet50, resnet101, resnet152
from .squeezenet import SqueezeNet, squeezenet1_0, squeezenet1_1
from .se_resnet import se_resnet18, se_resnet34, se_resnet50, se_resnet101, se_resnet152
from .darknet import darknet19, darknet53
from .mobilenet import MobileNetV1, MobileNetV2, MobileNetV3, MobileNetV3_Large, MobileNetV3_Small
from .densenet import DenseNet, densenet121, densenet161, densenet169, densenet201
from .mnasnet import MNASNet, mnasnet0_5, mnasnet0_75, mnasnet1_0, mnasnet1_3

# 2020s
from .conv2former import Conv2Former_n, Conv2Former_t, Conv2Former_s, Conv2Former_l, Conv2Former_b
from .ghostnet import GhostNet, GGhostRegNet, ghostnetv1, ghostnetv2, g_ghost_regnetx_002, \
    g_ghost_regnetx_004, g_ghost_regnetx_006, g_ghost_regnetx_008, g_ghost_regnetx_016, \
    g_ghost_regnetx_032, g_ghost_regnetx_040, g_ghost_regnetx_064, g_ghost_regnetx_080, \
    g_ghost_regnetx_120, g_ghost_regnetx_160, g_ghost_regnetx_320
from .regnet import regnetx_002, regnetx_004, regnetx_006, regnetx_008, regnetx_016, regnetx_032,\
    regnetx_040, regnetx_064, regnetx_080, regnetx_120, regnetx_160, regnetx_320
from .fasternet import fasternet_t0, fasternet_t1, fasternet_t2, fasternet_s, \
    fasternet_m, fasternet_l



"""
Copyright (c) 2024, Auorui.
All rights reserved.

Darknet-19: https://arxiv.org/pdf/1612.08242.pdf                     Yolov2
Darknet-53: https://pjreddie.com/media/files/papers/YOLOv3.pdf       Yolov3
Blog records: https://blog.csdn.net/m0_62919535/article/details/132639078
"""
import torch
import math
from torch import nn
from collections import OrderedDict

__all__ = ["darknet19", "darknet53"]

class ConvBNLR(nn.Sequential):
    """
    ConvBNLR -> Conv+BN+LeakyReLU
    """
    def __init__(self, in_channels, out_channels, kernel_size=3, stride=1, padding=1, bias=False, negative_slope=0.1, **kwargs):
        super(ConvBNLR, self).__init__(
            nn.Conv2d(in_channels, out_channels, kernel_size, stride, padding, bias=bias, **kwargs),
            nn.BatchNorm2d(out_channels),
            nn.LeakyReLU(negative_slope)
        )

class DarkResidualBlock(nn.Module):
    # Residual block
    def __init__(self, inplanes, planes):
        super(DarkResidualBlock, self).__init__()
        # block1 降通道数，block2 再将通道数升回去，如 64->32->64
        self.conv = nn.Sequential(
            ConvBNLR(inplanes, planes[0], kernel_size=1, padding=0),
            ConvBNLR(planes[0], planes[1])
        )

    def forward(self, x):
        residual = x
        out = self.conv(x)
        out += residual
        return out


class Darknet53(nn.Module):
    def __init__(self, num_classes, layers):
        super(Darknet53, self).__init__()
        self.inplanes = 32
        self.layer0 = ConvBNLR(3, self.inplanes)                    # (3, 224, 224)   -> (32, 224, 224)
        self.layer1 = self._make_layer([32, 64], layers[0])         # (32, 224, 224)  -> (64, 112, 112)
        self.layer2 = self._make_layer([64, 128], layers[1])        # (64, 112, 112)  -> (128, 56, 56)
        self.layer3 = self._make_layer([128, 256], layers[2])       # (128, 56, 56) -> (256, 28, 28)
        self.layer4 = self._make_layer([256, 512], layers[3])       # (256, 28, 28)   -> (512, 14, 14)
        self.layer5 = self._make_layer([512, 1024], layers[4])      # (512, 14, 14)   -> (1024, 7, 7)

        self.global_avg_pool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(1024, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    def forward(self, x):
        x = self.layer0(x)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        out = self.layer5(x)
        out = self.global_avg_pool(out)
        out = out.view(-1, 1024)
        out = self.fc(out)

        return out

    def _make_layer(self, planes, blocks):
        """
        在每一个layer里面，首先利用一个步长为2的3x3卷积进行下采样
        然后进行残差结构的堆叠
        """
        layers = []
        # 下采样，步长为2，卷积核大小为3
        layers.append(("down_sampling", ConvBNLR(self.inplanes, planes[1], kernel_size=3, stride=2, padding=1, bias=False)))
        self.inplanes = planes[1]
        for i in range(0, blocks):
            layers.append(("residual_{}".format(i), DarkResidualBlock(self.inplanes, planes)))
        return nn.Sequential(OrderedDict(layers))

class Darknet19(nn.Module):
    def __init__(self, num_classes, cfg):
        super(Darknet19, self).__init__()
        self.features = self.make_layers(cfg)
        self.global_avg_pool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(1024, num_classes)  # Assuming the last value before 'M' is the output channels

    def make_layers(self, cfg):
        """参考的VGG的复用结构"""
        layers = []
        in_channels = 3
        for i in cfg:
            if i == 'M':
                layers.append(nn.MaxPool2d(kernel_size=2, stride=2))
            else:
                out_channels, kernel_size, padding = i[0], i[1], i[2]
                layers.append(ConvBNLR(in_channels, out_channels, kernel_size=kernel_size, padding=padding))
                in_channels = out_channels
        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.features(x)
        x = self.global_avg_pool(x)
        x = x.view(x.size(0), -1)
        x = self.fc(x)
        return x

# Model structure reference from https://arxiv.org/pdf/1612.08242.pdf Table 6 —— Darknet19
#                                https://pjreddie.com/media/files/papers/YOLOv3.pdf Table 1 —— Darknet53
cfg = {"A":[(32, 3, 1), 'M', (64, 3, 1), 'M', (128, 3, 1), (64, 1, 0), (128, 3, 1), 'M', (256, 3, 1),
         (128, 1, 0), (256, 3, 1), 'M', (512, 3, 1), (256, 1, 0), (512, 3, 1), (256, 1, 0),
         (512, 3, 1), 'M', (1024, 3, 1), (512, 1, 0), (1024, 3, 1), (512, 1, 0), (1024, 3, 1)],

       "B":[1, 2, 8, 8, 4]}

def darknet19(num_classes):
    return Darknet19(num_classes, cfg["A"])

def darknet53(num_classes):
    return Darknet53(num_classes, cfg["B"])

if __name__=="__main__":
    import torchsummary
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    input = torch.ones(2, 3, 224, 224).to(device)
    net = darknet19(num_classes=4)
    net = net.to(device)
    out = net(input)
    print(out)
    print(out.shape)
    torchsummary.summary(net, input_size=(3, 224, 224))
    # Total params: 19,821,476  --19
    # Total params: 40,589,028  --53